'''
Functions for working with quaternions. Note that all the functions also
work on arrays, and can deal with full quaternions as well as with
quaternion vectors.

- deg2quat 
- rotate_vector
- quatinv
- quatmult
- quat2deg
- quat2vect
- vel2quat

'''

'''
author: Thomas Haslwanter
date:   Oct-2013
ver:    2.0
'''

from numpy import sqrt, sum, r_, c_, hstack, cos, sin, atleast_2d, zeros, shape, vstack, prod, min, arcsin, pi, tile
import matplotlib.pyplot as plt

def deg2quat(inDeg):
    '''Convert axis-angles or plain degree into the corresponding quaternion values.'''
    return sin(0.5 * inDeg * pi/180)
    
def rotate_vector(vector, q):
    vector = atleast_2d(vector)
    qvector = hstack((zeros((vector.shape[0],1)), vector))
    vRotated = quatmult(q, quatmult(qvector, quatinv(q)))
    vRotated = vRotated[:,1:]

    if min(vRotated.shape)==1:
        vRotated = vRotated.flatten()

    return vRotated


def quatinv(q):
    ''' Quaternion inversion '''
    q = atleast_2d(q)
    if q.shape[1]==3:
        return -q
    else:
        qLength = sum(q**2, 1)
        qConj = q * r_[1, -1,-1,-1]
        return (qConj.T / qLength).T

def quatmult(p,q):
    ''' Quaternion multiplication:
    Calculates the product of two quaternions r = p * q
    If one of both of the quaterions have only three columns,
    the scalar component is calculated such that the length
    of the quaternion is one.
    The lengths of the quaternions have to match, or one of
    the two quaternions has to have the length one.
    If both p and q only have 3 components, the returned quaternion
    also only has 3 components; otherwise it has all four.
    '''

    flag3D = False
    if atleast_2d(p).shape[1]==3 & atleast_2d(q).shape[1]==3:
        flag3D = True

    if len(p) != len(q):
        if len(p)==1 & len(q)==1:
            error('Both arguments in the quaternion multiplication must have the same number of rows, unless one has only one row.')

    p = vect2quat(p).T
    q = vect2quat(q).T
    
    if prod(shape(p)) > prod(shape(q)):
        r=zeros(shape(p))
    else:
        r=zeros(shape(q))

    r[0] = p[0]*q[0] - p[1]*q[1] - p[2]*q[2] - p[3]*q[3]
    r[1] = p[1]*q[0] + p[0]*q[1] + p[2]*q[3] - p[3]*q[2]
    r[2] = p[2]*q[0] + p[0]*q[2] + p[3]*q[1] - p[1]*q[3]
    r[3] = p[3]*q[0] + p[0]*q[3] + p[1]*q[2] - p[2]*q[1]

    if flag3D:
        # for rotations > 180 deg
        r[:,r[0]<0] = -r[:,r[0]<0]
        r = r[1:]

    r = r.T
    return r

def quat2deg(inQuat):
    '''Calculate the axis-angle corresponding to a given quaternion.'''
    return 2 * arcsin(quat2vect(inQuat)) * 180 / pi

def quat2vect(inQuat):
    '''Extract the quaternion vector from a quaternion.'''
    inQuat = atleast_2d(inQuat)
    if inQuat.shape[1] == 4:
        vect = inQuat[:,1:]
    else:
        vect = inQuat
    if min(vect.shape)==1:
        vect = vect.flatten()
    return vect

def vect2quat(inData):
    ''' Utility function, which turns a quaternion vector into a unit quaternion.'''
    inData = atleast_2d(inData)
    (m,n) = inData.shape
    if (n!=3)&(n!=4):
        error('Quaternion must have 3 or 4 columns')
    if n == 3:
        qLength = 1-sum(inData**2,1)
        numLimit = 1e-12
        # Check for numerical problems
        if min(qLength) < -numLimit:
            error('Quaternion is too long!')
        else:
            # Correct for numerical problems
            qLength[qLength<0] = 0
        inData = hstack((c_[sqrt(qLength)], inData))
    return inData

def vel2quat(vel, q0, rate, CStype):
    '''
    Take an angular velocity (in deg/s), and convert it into the
    corresponding orientation quaternion.
    q0 ... vector-part of quaternion (!!)
    '''
    
    # Convert from deg/s to rad/s
    vel = vel * pi/180
    vel_t = sqrt(sum(vel**2, 1))
    vel_nonZero = vel_t>0
    
    # initialize the quaternion
    q_delta = zeros(shape(vel))
    q_pos = zeros((len(vel),4))
    q_pos[0,:] = vect2quat(q0)
    
    # magnitude of position steps
    dq_total = sin(vel_t[vel_nonZero]/(2*rate))
    
    q_delta[vel_nonZero,:] = vel[vel_nonZero,:] * tile(dq_total/vel_t[vel_nonZero], (3,1)).T
    
    for ii in range(len(vel)-1):
        q1 = vect2quat(q_delta[ii,:])
        q2 = q_pos[ii,:]
        if CStype == 'sf':            
            qm = quatmult(q1,q2)
        elif CStype == 'bf':
            qm = quatmult(q2,q1)
        else:
            print 'I don''t know this type of coordinate system!'
        q_pos[ii+1,:] = qm
    
    return q_pos

if __name__=='__main__':
    '''These are some simple tests to see if the functions produce the
    proper output.'''
    a = r_[cos(0.1), 0,0,sin(0.1)]
    b = r_[cos(0.1),0,sin(0.1), 0]
    c = vstack((a,b))
    d = r_[sin(0.1), 0, 0]
    e = r_[2, 0, sin(0.1), 0]

    print quatmult(a,a)
    print quatmult(a,b)
    print quatmult(c,c)
    print quatmult(c,a)
    print quatmult(d,d)

    print('The inverse of {0} is {1}'.format(a, quatinv(a)))
    print('The inverse of {0} is {1}'.format(d, quatinv(d)))
    print('The inverse of {0} is {1}'.format(e, quatinv(e)))
    print quatmult(e, quatinv(e))

    print quat2vect(a)
    print('{0} is {1} degree'.format(a, quat2deg(a)))
    print('{0} is {1} degree'.format(c, quat2deg(c)))
    print quat2deg(0.2)
    x = r_[1,0,0]
    vNull = r_[0,0,0]
    print rotate_vector(x, a)

    v0 = [0., 0., 100.]
    vel = tile(v0, (1000,1))
    rate = 100
    
    out = vel2quat(vel, [0., 0., 0.], rate, 'sf')
    print out[-1:]
    plt.plot(out[:,1:4])
    plt.show()
    
    print deg2quat(15)
    print deg2quat(quat2deg(a))
    
